const fs = require('fs-extra')
const path = require('path')
const ora = require('ora')
const kebabCase = require('lodash.kebabcase')
const { print, run } = require('@toptal/davinci-cli-shared')

const skeletonLibMainPath = require.resolve('@toptal/davinci-skeleton')

const NODE_MODULES = /skeleton\/node_modules/g
const CHANGELOG = /CHANGELOG.md/g

const checkPrerequisites = async (currentRunningDir, appName) => {
  const spinner = ora('Checking prerequisites...').start()

  // look for env resolvers
  try {
    await run('direnv', ['version'])
  } catch (err) {
    spinner.warn()
    print.yellow(
      "warning: You don't have `direnv` for resolving env variables! Expect other approach is used."
    )
  }

  // check if app already exists
  const appPath = path.join(currentRunningDir, appName)
  const doesAppExists = await fs.exists(appPath)

  if (doesAppExists) {
    spinner.fail()
    throw new Error(`Application with name ${appName} already exists!`)
  }

  spinner.succeed('Prerequisites checked!')
}

const copyApplicationSkeleton = async (dest, appName) => {
  const appDir = path.join(skeletonLibMainPath, '../..')
  const destDir = path.join(dest, appName)

  const spinner = ora('Copying skeleton...').start()

  const filterSkeletonFiles = src =>
    !src.match(NODE_MODULES) && !src.match(CHANGELOG)

  try {
    await fs.copy(appDir, destDir, { filter: filterSkeletonFiles })

    // prepare package.json
    const packageJsonFile = path.join(destDir, 'package.json')
    const packageSkeletonJsonFile = path.join(destDir, 'package.json.skeleton')

    if (fs.existsSync(packageSkeletonJsonFile)) {
      await fs.remove(packageJsonFile)
      await fs.rename(packageSkeletonJsonFile, packageJsonFile)
    }

    // prepare .gitignore
    const gitIgnoreFile = path.join(destDir, '.gitignore')
    const gitIgnoreSkeletonFile = path.join(destDir, '.gitignore.skeleton')

    if (fs.existsSync(gitIgnoreSkeletonFile)) {
      await fs.rename(gitIgnoreSkeletonFile, gitIgnoreFile)
    }

    // remove bin directory
    await fs.remove(path.join(destDir, 'bin'))

    spinner.succeed('Skeleton copied!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const modifyPackageJson = async (dest, appName) => {
  const packageJsonFile = path.join(dest, appName, './package.json')

  const spinner = ora('Modifiying package.json...').start()

  try {
    const configData = fs.readFileSync(packageJsonFile)
    const config = JSON.parse(configData)

    config.name = `@toptal/${appName}`
    config.homepage = `https://github.com/toptal/${appName}/tree/master#readme`

    delete config.gitHead

    delete config.scripts['prepack']
    delete config.scripts['postpack']
    delete config.scripts['build:package']
    delete config.scripts['prepublishOnly']

    delete config.publishConfig
    config.license = 'SEE LICENSE IN LICENSE.MD'

    config.version = '0.0.1'

    const modifiedConfigData = JSON.stringify(config, null, 2).concat('\n')

    fs.writeFileSync(packageJsonFile, modifiedConfigData)

    spinner.succeed('package.json modified!')
  } catch (err) {
    spinner.fail()
    throw err
  }
}

const installDependencies = async (dest, appName) => {
  const appDir = path.join(dest, appName)

  const spinner = ora('Installing dependencies... ').start()

  // list of dependencies to be added to a new project
  const dependencies = [
    '@toptal/picasso-provider',
    '@toptal/picasso',
    'react',
    'react-dom',
    'styled-components',
    'typescript',
    '@apollo/client',
    'graphql'
  ]

  const devDependencies = [
    '@toptal/davinci',
    '@types/react',
    '@types/react-dom'
  ]

  try {
    await run('yarn', ['add', ...dependencies], { cwd: appDir })
    await run('yarn', ['add', '-D', ...devDependencies], { cwd: appDir })
    spinner.succeed('Dependencies installed!')
  } catch (err) {
    spinner.fail()
    throw new Error(err)
  }
}

const createCommand = async appName => {
  print.green('Creating', appName, '...')

  const currentRunningDir = process.cwd()

  try {
    // check if all apps are installed, directory doesn't exists
    await checkPrerequisites(currentRunningDir, appName)

    // copy skeleton to destination directory
    await copyApplicationSkeleton(currentRunningDir, appName)

    // change package.json
    await modifyPackageJson(currentRunningDir, appName)

    // run yarn
    await installDependencies(currentRunningDir, appName)

    print.green('Application created.')
  } catch (err) {
    print.red(err)
    process.exit(1)
  }
}

const newCommandCreator = {
  command: 'new <appName>',
  description: 'Create new application',
  action: async ({ appName }) => {
    const formattedAppName = kebabCase(appName)

    return createCommand(formattedAppName)
  }
}

module.exports = newCommandCreator
